/*
   ScanJet Note Assembler 
   (c) 2005 by GanjaTron (ganjatron@gmx.net)
   
   Generates binary song files for HP ScanJet from simple ASCII input from
   stdin.

   Input files have the following text format:
   
   <tempo> 
   % Dees ees a comment, babee
   <note1> <note2> <note3> ...
   ...
   ...,
   
   where tempo is expressed in beats/sec (possibly fractional, max 8). 
   Anything following '%' is treated as a comment and ignored. Notes have
   the following format:
   
   <noteN> = <note><octave>[duration],
   
   where:
   
   -- Note is one of {C, C#, D, D#, E, F, F#, G, G#, A, A#, B, P}, 
      where P denotes a pause.
   -- Octave is in the range [0-2], with C1 denoting middle C. Octaves 
      are omitted for pauses.
   -- Duration is in number of beats (possibly fractional), or one beat 
      if omitted.
   
   Notes must lie in the ScanJet's tonal range D#0 -- A3 (ca. 3 octaves).

   -----------------------------------------------------------------------
   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; use at your own risk. The author assumes no
   liability for any damages incurred by its use, particularly to scanning
   hardware. Care has been taken to avoid potential damage, but if your
   scanner makes strange noises SHUT IT DOWN IMMEDIATELY!.

   By using this software, you agree to these terms and conditions.      
   -----------------------------------------------------------------------
*/



/* Scanner's stepping motor clock frequency in Hz */
#define STEPCLK 1.5e6
/* Minimum duration in sec */
#define MINDURATION (1.0 / 8)
/* Tonal range covers 3 octaves (D#0 -- A3) */
#define MINFREQ (freq [3] * 0.5)
#define MAXFREQ (freq [9] * 4)



#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>



/* Note frequency table [Hz] starting at middle C.
   Taken from: Chamberlin, "A Sampling of Techniques for Computer
   Performance of Music", BYTE, sept. 1977. */
const float freq [] = {261.62, 277.18, 293.66, 311.13, 329.63, 349.23, 
                       369.99, 391.99, 415.30, 440.00, 466.16, 493.28, 0};
const char* notes = "C C#D D#E F F#G G#A A#B P";



int getUpperStr (FILE *file, char* str, unsigned *lineCnt)
/* Gets space-delimited string from file and converts to uppercase.
   Line counter is incremented when string is at beginning of line. */
{
  char c, *s = str;
  int comment = 0;
  
  for (;;) {
     c = fgetc(file); 
     if (feof(file)) break;
     comment |= c == '%';
     /* Stuff in comments is ignored */
     if (!comment) {
        /* Skip leading space */
        if (!isspace(c)) *s++ = toupper(c);
        else if (s != str) {
           /* Found trailing space; push back newline 
              for next read if at end of line */
           if (c == '\n') ungetc(c, file);
           break;
        }
     }
     if (c == '\n') {
        /* New line; increment line counter & reset comment flag */
        (*lineCnt)++;
        comment = 0;
     }
  }
  *s = '\0';
  return (s != str);
}



void scanJetNoteAsm (FILE *inFile, FILE *outFile)
{
   float f;
   unsigned tempo, note, oct, duration, lineCnt = 1;
   char s1 [80], s2 [80], *p1, *p2;
   
   /* Parse tempo (num of 1/8s cycles per beat) */
   if (!getUpperStr(inFile, s1, &lineCnt) || !(f = atof(s1))) {
      fprintf(stderr, "Line %d: Invalid tempo\n", lineCnt);
      exit(1);
   }
   tempo = 1 / (f * MINDURATION) + 0.5;
   if (tempo < 1 || tempo >= 1 << 8) {
      fprintf(stderr, "Line %d: Tempo out of range\n", lineCnt);
      exit(1);
   }
   /* Parse notes */
   while (getUpperStr(inFile, s1, &lineCnt)) {
      /* Extract note */
      for (p1 = s1, p2 = s2; !isspace(*p1) && strchr(notes, *p1); 
           *p2++ = *p1++);
      *p2 = '\0';
      if (!*s2 || !(p2 = strstr(notes, s2))) {
         fprintf(stderr, "Line %d: Invalid note %s\n", lineCnt, s2);
         exit(1);
      }
      note = p2 - notes >> 1;
      f = freq [note];
      if (f) {
         /* Extract octave (skipped for pause) */
         oct = *p1++ - '0';
         f *= 0.5 * (1 << oct);
         if (f < MINFREQ || f > MAXFREQ) {
            fprintf(stderr, "Line %d: Note %s%d out of range\n", 
                    lineCnt, s2, oct);
            exit(1);
         }
         /* Number of 1.5 MHz clock cycles between scanner motor steps */
         note = STEPCLK / f + 0.5;
      }
      else note = 0;
      /* Extract duration (1 if omitted) */
      for (p2 = s2; isdigit(*p1) || *p1 == '.'; *p2++ = *p1++);
      *p2 = '\0';
      if (!*s2) f = 1;
      else if (!(f = atof(s2))) {
         fprintf(stderr, "Line %d: Invalid duration %s\n", lineCnt, s2);
         exit(1);
      }
      if ((duration = f * tempo + 0.5) >= 1 << 8) {
         fprintf(stderr, "Line %d: duration out of range\n", lineCnt);
         exit(1);
      }
      /* Write to binary output file */
      s1 [0] = note >> 8;
      s1 [1] = note;
      s1 [2] = duration;
      note = fwrite(s1, 1, 3, outFile);
      if (note != 3) {
         fputs("Error writing to output file\n", stderr);
         exit(1);
      }
   }
}



int main ()
{
   scanJetNoteAsm(stdin, stdout);
   return 0;
}
